/*
 * file:makeUtf16ToEucJpTable.c
 */
#define STRICT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*//////////////////////////////////////////////////////////////////*/
/*
 * UTF16 → EUC-JP マッピング表(入力ファイル名)
 */
#define INPFILENAME "Unicode_Euc_Jp_mapping.txt"

/*//////////////////////////////////////////////////////////////////*/
/*
 * UTF16 → EUC-JP マッピング表(出力ファイル名)
 */
#define OUTFILENAME "utf16_to_euc_jp_table.h"

/*//////////////////////////////////////////////////////////////////*/
/*
 * ダミーコード。
 * 対応するコードが存在しない場合に戻すコード
 */
#define DUMMY_CODE 0x0000

/*//////////////////////////////////////////////////////////////////*/
/*
 * ファイル出力時に、１行に何件のデータを出力するかを指定。
 */
#define COLUMNS_PER_LINE 2

/*//////////////////////////////////////////////////////////////////*/
/*
 * BITMAP_INDEX構造体
 * UTF-16コードよりEUC-JPコードへの対応を検索するためのインデックスの構造
 */
struct _BITMAP_INDEX
{
    unsigned char byType;           /* 1:isEmpty 2:BitmapIndex 3:Data */
    unsigned long dwBitmapIndex;    /* BitmapIndexへのポインタ(0の場合はBitmapIndexが存在しない) */
    unsigned long dwEucJpCode;      /* EUC_JP コード(未知の場合には0) */
    unsigned long dwUtf16Code;      /* UTF16コード(未知の場合には0) */
};

/*//////////////////////////////////////////////////////////////////*/
/*
 * UTF-16 → EUC-JP マッピングデータの構造体
 * ファイルより読み込んだデータを処理するために双方向リスト構造へ格納。
 */
struct _EUC_JP_UNICODE_MAPPING_DATA
{
    unsigned long dwRow;            /* 行番号 */
    unsigned long dwEucJpCode;      /* EUC_JP コード(未知の場合には0) */
    unsigned long dwUtf16Code;      /* UTF16コード(未知の場合には0) */
    struct _EUC_JP_UNICODE_MAPPING_DATA *prev;
    struct _EUC_JP_UNICODE_MAPPING_DATA *next;
} *g_head = 0, *g_tail = 0;

/*//////////////////////////////////////////////////////////////////*/
/*
 * 構造体の typedef 宣言
 */
typedef struct _BITMAP_INDEX T_BITMAP_INDEX;
typedef struct _EUC_JP_UNICODE_MAPPING_DATA T_EUC_JP_UNICODE_MAPPING_DATA;

/*//////////////////////////////////////////////////////////////////*/
/*
 * 読み込むデータ領域の大きさを指定
 */
#define g_size_of_data (20000)

/*//////////////////////////////////////////////////////////////////*/
/*
 * インデックスを作成する領域の大きさを指定
 */
#define g_size_of_index (256 * 256 * 256)

/*//////////////////////////////////////////////////////////////////*/
/*
 * インデックス
 */
T_BITMAP_INDEX g_index[g_size_of_index];

/*//////////////////////////////////////////////////////////////////*/
/*
 * インデックスのデータを格納する最終位置
 * これより以降のデータはまだ未使用である事を意味する
 */
unsigned long g_tail_index = 256;

/*//////////////////////////////////////////////////////////////////*/
/*
 * 入力ファイルへのポインタ
 */
FILE *pInpFile = NULL;

/*//////////////////////////////////////////////////////////////////*/
/*
 * 関数の宣言部
 */
int alterIndex();
int freeData();
int freeIndex();
int insertData(unsigned long eucJpCode, unsigned long unicode);
int makeUtf16ToEucJpTable();
int outprefix(FILE *pFile);
int outputDataToFile(FILE* pOutFile);
int outsuffix(FILE *pFile);
int printData();
int printIndex();
unsigned long readRow(unsigned long eucJpcode);

/*//////////////////////////////////////////////////////////////////*/
/*
 * main 関数
 * ここからプログラムの実行が開始されます。
 */
int main(int argc, char *argv[])
{
    int iResult = 0;
    
    iResult = makeUtf16ToEucJpTable();
    if (iResult)
    {
        fprintf(stderr, "makeUtf16ToEucJpTable Failed: %d\n", iResult);
    }
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * alterIndex() 関数
 * インデックスを再構築します。
 */
int alterIndex()
{
    T_EUC_JP_UNICODE_MAPPING_DATA *data;
    unsigned long   unicode         = 0;
    unsigned long   eucJpCode       = 0;
    unsigned long   firstByte       = 0;
    unsigned long   secondByte      = 0;
    unsigned long   firstIndex      = 0;
    unsigned long   secondIndex     = 0;
    
    freeIndex();
    
    data = g_head;
    
    while (data)
    {
        eucJpCode = data->dwEucJpCode;
        unicode = data->dwUtf16Code;
        
        firstByte = (unicode >> 8) & 0x0000ff;
        secondByte = (unicode) & 0x0000ff;
        firstIndex = firstByte;
        if (g_index[firstIndex].byType != 2)
        {
            g_index[firstIndex].byType = 2;
            g_index[firstIndex].dwBitmapIndex = g_tail_index;
            g_index[firstIndex].dwEucJpCode = 0;
            g_index[firstIndex].dwUtf16Code = 0;
            g_tail_index += 256;
        }
        secondIndex = g_index[firstIndex].dwBitmapIndex + secondByte;
        g_index[secondIndex].byType = 3; /* 1:isEmpty 2:BitmapIndex 3:Data */
        g_index[secondIndex].dwBitmapIndex = 0;
        g_index[secondIndex].dwEucJpCode =eucJpCode;
        g_index[secondIndex].dwUtf16Code = unicode;
        
        data = data->next;
    }
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * freeData() 関数
 * 双方向リストのデータを解放します。
 */
int freeData()
{
    T_EUC_JP_UNICODE_MAPPING_DATA *data;
    
    while (g_tail)
    {
        data = g_tail;
        g_tail = data->prev;
        free(data);
    }
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * freeIndex() 関数
 * インデックス情報を開放します。
 */
int freeIndex()
{
    size_t count = 0;
    
    count = sizeof(T_BITMAP_INDEX) * g_size_of_index;
    memset(g_index, 0, count);
    g_tail_index = 256;
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * insertData() 関数
 * データを一件挿入します。
 */
int insertData(unsigned long eucJpCode, unsigned long unicode)
{
    T_EUC_JP_UNICODE_MAPPING_DATA *data;
    size_t size = 0;
    
    size = sizeof(T_EUC_JP_UNICODE_MAPPING_DATA);
    data = (T_EUC_JP_UNICODE_MAPPING_DATA *)malloc(size);
    data->dwEucJpCode = eucJpCode;
    data->dwUtf16Code = unicode;
    data->next = 0;
    data->prev = 0;
    if (g_head == 0)
    {
        data->dwRow = 0;
        g_head = data;
        g_tail = data;
    }
    else
    {
        data->dwRow = g_tail->dwRow + 1;
        data->prev = g_tail;
        g_tail->next = data;
        g_tail = data;
    }
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * makeUtf16ToEucJpTable() 関数
 * UTF-16 → EUC-JP 変換テーブルを作成します
 */
int makeUtf16ToEucJpTable()
{
    unsigned long   unicode     = 0;
    unsigned long   eucJpCode   = 0;
    FILE* pOutFile = NULL;
    
    /* ●データの読み込み */
    
    /*
     * 制御コード(0x0000～0x001f)
     */
    for (unicode = 0x0000; unicode <= 0x001f; unicode++)
    {
        eucJpCode = readRow(unicode);
        insertData(eucJpCode, unicode);
    }
    
    /*
     * ASCII文字(0x0020～0x007e)
     */
    
    for (unicode = 0x0020; unicode <= 0x007e; unicode++)
    {
        eucJpCode = readRow(unicode);
        insertData(eucJpCode, unicode);
    }
    
    /*
     * 制御コード(0x007f)
     */
    for (unicode = 0x007f; unicode <= 0x007f; unicode++)
    {
        eucJpCode = readRow(unicode);
        insertData(eucJpCode, unicode);
    }
    
    /*
     * 漢字(0x0080～0xfffe)
     */
    for (unicode = 0x0080; unicode <= 0xfffe; unicode++)
    {
        eucJpCode = readRow(unicode);
        if (eucJpCode != DUMMY_CODE) {
            insertData(eucJpCode, unicode);
        }
    }
    
    /* ●インデックスを張る */
    alterIndex();

    /* ●ファイルを出力 */
    pOutFile = fopen(OUTFILENAME, "w");
    if (pOutFile != NULL)
    {
        outprefix(pOutFile);
        outputDataToFile(pOutFile);
        outsuffix(pOutFile);
    }
    
    /* ●ファイルを閉じる */
    if (pInpFile != NULL)
    {
        fclose(pInpFile);
        pInpFile = NULL;
    }
    
    if (pOutFile != NULL)
    {
        fclose(pOutFile);
        pOutFile = NULL;
    }
    
    /* ●データを解放する */
    freeData();
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * outprefix() 関数
 * プレフィックスをファイルへ出力
 */
int outprefix(FILE *pFile)
{
    fprintf(pFile, 
"/*\n\
 * file:%s\n\
 */\n\
/*\n\
# Unicode EUC-JP\n\
# Character Code Mappings\n\
#   <EUC-JP>\n\
#       <制御コード>0x00～0x1f, 0x7f</制御コード>\n\
#       <ASCII文字>0x20～0x7e</ASCII文字>\n\
#       <半角カタカナ>0x8ea1～0x8edf</半角カタカナ>\n\
#       <漢字>0xa1a1～0xf4a6, 0x8fa2af～0x8fede3<漢字>\n\
#   </EUC-JP>\n\
*/\n"
            , OUTFILENAME);
    fprintf(pFile, 
"struct _BITMAP_INDEX\n\
{\n\
    unsigned char byType;           /* 1:isEmpty 2:BitmapIndex 3:Data */\n\
    unsigned long dwBitmapIndex;    /* BitmapIndexへのポインタ(0の場合はBitmapIndexが存在しない) */\n\
    unsigned long dwEucJpCode;      /* EUC_JP コード(未知の場合には0) */\n\
    unsigned long dwUtf16Code;      /* UTF16コード(未知の場合には0) */\n\
};\n\
typedef struct _BITMAP_INDEX T_BITMAP_INDEX;\n\
\n\
const T_BITMAP_INDEX utf16_to_euc_jp_table[] = {\n\
"
            );
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * outputDataToFile() 関数
 * テーブル情報をファイルへ出力
 */
int outputDataToFile(FILE* pOutFile)
{
    unsigned long row = 0;
    unsigned long last = 0;
    
    if (pOutFile == NULL)
    {
        return 1;
    }
    
    last = g_tail_index;
    
    if (last < 255)
    {
        last = 255;
    }
    
    for (row = 0; row <= last; row++)
    {
        if (g_index[row].byType == 0)
        {
            g_index[row].byType = 1;    /* 1:isEmpty 2:BitmapIndex 3:Data */
            g_index[row].dwBitmapIndex = 0;
            g_index[row].dwEucJpCode = 0;
            g_index[row].dwUtf16Code = 0;
            
        }
        
        if (row == 0)
        {
            fprintf(pOutFile, "\t");
        }
        else if ((row % COLUMNS_PER_LINE) == 0)
        {
            fprintf(pOutFile, ",\n\t");
        }
        else
        {
            fprintf(pOutFile, ", ");
        }
        
        /* byType, dwBitmapIndex, dwEucJpCode, dwUtf16Code */
        fprintf(pOutFile, "{%d, 0x%08lX, 0x%06lX, 0x%04lX}"
                , g_index[row].byType
                , g_index[row].dwBitmapIndex
                , g_index[row].dwEucJpCode
                , g_index[row].dwUtf16Code);
    }
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * outsuffix() 関数
 * サーフィックスをファイルへ出力
 */
int outsuffix(FILE *pFile)
{
    fprintf(pFile, 
"\n\
};\n");
    
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * printData() 関数
 * データをコンソールへ出力
 */
int printData()
{
    T_EUC_JP_UNICODE_MAPPING_DATA *data;
    
    data = g_head;
    
    while (data)
    {
        printf("{0x%08lX, 0x%06lX, 0x%04lX}\n", data->dwRow, data->dwEucJpCode, data->dwUtf16Code);
        
        data = data->next;
    }
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * printIndex() 関数
 * インデックスをコンソールへ出力
 */
int printIndex()
{
    unsigned long row = 0;
    unsigned long last = 0;
    
    last = g_tail_index;
    
    if (last < 255)
    {
        last = 255;
    }
    
    for (row = 0; row <= last; row++)
    {
        if (g_index[row].byType != 0)
        {
            printf("{row:0x%08lX, byType:%d, dwBitmapIndex:0x%08lX, dwEucJpCode:0x%06lX, dwUtf16Code:0x%04lX}\n"
                    , row
                    , g_index[row].byType
                    , g_index[row].dwBitmapIndex
                    , g_index[row].dwEucJpCode
                    , g_index[row].dwUtf16Code);
        }
    }
    
    return 0;
}

/*//////////////////////////////////////////////////////////////////*/
/*
 * readRow() 関数
 * データを入力ファイルより読み出す
 */
unsigned long readRow(unsigned long eucJpcode)
{
    char string[512];
    int i = 0;
    unsigned long a;
    unsigned long b;
    long pos = 0;
    
    if (pInpFile == NULL)
    {
        pInpFile = fopen(INPFILENAME, "rb");
        if (pInpFile == NULL)
        {
            return DUMMY_CODE;
        }
    }
    
    for (i = 0; i < 10000; i++)
    {
        memset(string, 0, sizeof(string));
        pos = ftell(pInpFile);
        if (fgets(string, sizeof(string) - 1, pInpFile) == NULL)
        {
            fseek(pInpFile, 0, SEEK_SET);
            continue;
        }
        if (*string == '#' || *string == '\0')
        {
            continue;
        }
        
        a = 0;
        b = 0;
        sscanf(string, "%lx\t%lx", &a, &b);
        if (a == eucJpcode)
        {
            return b;
        }
        else
        {
            fseek(pInpFile, pos, SEEK_SET);
            return DUMMY_CODE;
        }
    }
    
    return DUMMY_CODE;
}
